#include "common.h"

//double Get_distance(LOCATION, LOCATION);

double Get_LCS_theta(double alpha, double beta, double gamma, double GCS_theta, double GCS_pi);
double Get_LCS_pi(double alpha, double beta, double gamma, double GCS_theta, double GCS_pi);


void LINK::Configuration(int _self_ms_idx)
{
	
	self_ms_idx= _self_ms_idx;

	Get_signal_interference();
	/*
	INPUT: pathloss, antgain
	OUTPUT: Signal power, interference power
	Notes: Large Scale  SINRϱ  interference  κ
	*/

	Get_adj_SECTORS();
	/*
	INPUT: static gain
	OUTPUT: adj_sector
	Notes:  ms   index 3 ã
	*/

	UE_Initial_Setting();
	/*
	INPUT:
	OUTPUT: incar_loss, power_per_sc, UE_ICI, UE_PATH_LOSS
	Notes: (outdoor to car loss)
	*/

	SINR = 10. * log10(pow(10., signal / 10.) / (pow(10., interference / 10.) + pow(10., (noise + MS_noisefig) / 10.)));

}



void LINK::Get_signal_interference()
{

	double strongest_signal;
	double pathloss;
	double antgain;
	double RMS_delay;
	double AOA_spread;
	double AOD_spread;
	interference = 0.;

	if (TYPE == 11) //// 5G InH
	{
		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
		{

			//pathloss = channel[ms[self_ms_idx].nearest_bs_idx][self_ms_idx].pathloss; //channel  pathloss´ (⼭ pathloss pathloss + shadowing)
			pathloss = channel[bs_idx][self_ms_idx].pathloss;
			RMS_delay = channel[bs_idx][self_ms_idx].RMS_delay_spread;
			AOA_spread = channel[bs_idx][self_ms_idx].circular_angle_spread_AOA;
			AOD_spread = channel[bs_idx][self_ms_idx].circular_angle_spread_AOD;

			///antgain = 0.;   //omnidirectional case


			antgain = Get_antgain(channel[bs_idx][self_ms_idx].self_bs, channel[bs_idx][self_ms_idx].self_ms, 0);   ///// MAX antgain - combined gain

			signal = bs_maxpower + antgain - pathloss;

			if (bs_idx == 0)
			{
				strongest_signal = signal;
				self_bs_idx = ms[self_ms_idx].nearest_bs_idx;
				self_sector_idx = 0;
				link_pathloss = pathloss;
				link_antgain = antgain;
				link_RMS_delay_spread = RMS_delay;
				link_AOA_spread = AOA_spread;
				link_AOD_spread = AOD_spread;

				// indoor num_BS = num_sector , bs = sector ׳  

				_sector_in_control = self_bs_idx;

			}
			else if (signal > strongest_signal)
			{
				interference += pow(10., strongest_signal / 10.);
				strongest_signal = signal;
				self_bs_idx = ms[self_ms_idx].nearest_bs_idx;
				self_sector_idx = 0;
				link_pathloss = pathloss;
				link_antgain = antgain;
				link_RMS_delay_spread = RMS_delay;
				link_AOA_spread = AOA_spread;
				link_AOD_spread = AOD_spread;


				_sector_in_control = self_bs_idx;

			}
			else
			{
				interference += pow(10., signal / 10.);
			}
		}

	}

	else
	{
		
		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
		{

			pathloss = channel[bs_idx][self_ms_idx].pathloss; //channel  pathloss´ (⼭ pathloss pathloss + shadowing) // self_ms_idx   19BS  
			RMS_delay = channel[bs_idx][self_ms_idx].RMS_delay_spread;
			AOA_spread = channel[bs_idx][self_ms_idx].circular_angle_spread_AOA;
			AOD_spread = channel[bs_idx][self_ms_idx].circular_angle_spread_AOD;

			for (int sector_idx = 1; sector_idx <= 3; sector_idx++)
			{
	
				antgain = Get_antgain(channel[bs_idx][self_ms_idx].self_bs, channel[bs_idx][self_ms_idx].self_ms, sector_idx);   ///// MAX antgain - combined gain

				//antgain = 0.; // for small scale calibration
				

				signal = bs_maxpower + antgain - pathloss;

				if (bs_idx == 0 && sector_idx == 1)
				{
					strongest_signal = signal;
					self_bs_idx = bs_idx;
					self_sector_idx = sector_idx;
					link_pathloss = pathloss;
					link_antgain = antgain;
					serving_bs = channel[bs_idx][self_ms_idx].self_bs;
					link_RMS_delay_spread = RMS_delay;
					link_AOA_spread = AOA_spread;
					link_AOD_spread = AOD_spread;

					_sector_in_control = self_bs_idx * 3 + (self_sector_idx - 1);


				}
				else if (signal > strongest_signal)
				{

					interference += pow(10., strongest_signal / 10.);
					strongest_signal = signal;
					self_bs_idx = bs_idx; // serving BS idx
					self_sector_idx = sector_idx;
					link_pathloss = pathloss;
					link_antgain = antgain;
					serving_bs = channel[bs_idx][self_ms_idx].self_bs;
					link_RMS_delay_spread = RMS_delay;
					link_AOA_spread = AOA_spread;
					link_AOD_spread = AOD_spread;

					_sector_in_control = self_bs_idx * 3 + (self_sector_idx - 1);


				}
				else
				{
					interference += pow(10., signal / 10.);
				}
			}
		}
	}

	str_signal = strongest_signal;
	signal = strongest_signal;
	interference = 10. * log10(interference);

}

void LINK::Get_adj_SECTORS() //num_sector ε ڿ ߾ static ,  ã
{
	LS_gain = new double[num_SECTORS];      //LS gain antgain - (pathloss + shadowing) sorting Ѱ
	static_gain = new double[num_SECTORS];  //antgain - (pathloss + shadowing)
	adj_sector = new int[num_compute_coef]; //coeff ŭ sector ε 

	double pathlosss;
	double antgainn;

	
	if (TYPE == 11) //// 5G InH
	{
		//NO USE
	}

	else
	{
		/////////////////////////////////////
		///LS gain GET
		/////////////////////////////////////
		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
		{
			pathlosss = channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].pathloss; //channel  pathloss´ (⼭ pathloss pathloss + shadowing) //

			for (int sector_idx = 1; sector_idx <= 3; sector_idx++)
			{

				antgainn = Get_antgain(channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].self_bs, channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].self_ms, sector_idx);

				//signal = bs_maxpower + antgain - pathloss;

				LS_gain[ wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx] * 3 + (sector_idx - 1)] = antgainn - pathlosss; //sorting Ѱ
				static_gain[ wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx] * 3 + (sector_idx - 1)] = antgainn - pathlosss; //static gain num_SECTORS  ߾ Ѱ
			}
		}
		///////////////////////////////////////////////////////

		///////////////////////////////////
		//LS gain sorting  ///////////////
		//////////////////////////////////
		for (int i = 1; i < num_SECTORS; i++)
		{
			double temp;
			int j;

			temp = LS_gain[i]; //LSgain antgain - (shadowing + pathloss)
			for (j = i; j > 0; j--)
			{
				if (LS_gain[j - 1] < temp)
					LS_gain[j] = LS_gain[j - 1];
				else break;
			}
			LS_gain[j] = temp;
		} //LS_gain ūź  ۰ sorting, [0]  ũ []  , LS_gain[0]  ū pathloss   -> 
		//////////////////////////////////////


		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
		{
			pathlosss = channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].pathloss; //channel  pathloss´ (⼭ pathloss pathloss + shadowing)

			for (int sector_idx = 1; sector_idx <= 3; sector_idx++)
			{
				antgainn = Get_antgain(channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].self_bs, channel[wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx]][self_ms_idx].self_ms, sector_idx);
				///////////////////////////////////////////////////////////////////////////////////////////////////////
				for (int idx = 0; idx < num_compute_coef; idx++)
				{
					if ((int)(1000000 * LS_gain[idx]) == (int)(1000000 * (antgainn - pathlosss)))
					{
						adj_sector[idx] = wrap_mat[ms[self_ms_idx].nearest_bs_idx][bs_idx] * 3 + (sector_idx - 1); //ϴ sector_idx ãƼ  // bs_idx*3 + (sector_idx - 1) ϸ num_SECTORS .
					}
				}
				//_sector_in_control = adj_sector[0]; //[0]    = sector_IN_control
				//////////////////////////////////////////////////////////////////////////////////////////////////////
			}
		}

		//////////   sector sector_in_control ϵ TEST /////////////

		
		//int idx = self_ms_idx / 10;

		//_sector_in_control = idx;

		//cout << "_sector_in_control = " << _sector_in_control << endl;
		//getchar();
		
		/////////////////////////////////////////////////////////////////////////////////////////////

	}
}


void LINK::UE_Initial_Setting(void) 
{


	//// Car penetration loss  //////////////////////////////////////////

  if (TYPE == 12) // Dense Urban 
  {
	  if (ms[self_ms_idx].Indoor == 1)   ///// Indoor
	  {
		  incar_loss = 0.;
	  }
	  else                              ////// outdoor
	  {
		  incar_loss = 9. + 5. * randnum.n();
	  }

  }
  else if (TYPE == 13) // Rural
  {
	  if (ms[self_ms_idx].Indoor == 1)   ///// Indoor
	  {
		  incar_loss = 0.;
	  }
	  else                              ////// outdoor
	  {
		  if (Configuration_Type == 2)     /////// Rural configuration C = 40% outdoor pedestrian, 20% outdoor incar
		  {
			  if (randnum.u() < (0.4/0.6))  //// Pedestrian
			  {
				  incar_loss = 0.;
			  }
			  else                   //// In Car
			  {
				  incar_loss = 9. + 5. * randnum.n();
			  }
		  }
		  else    /// 50% All outdoor
		  {
			  incar_loss = 9. + 5. * randnum.n();
		  }
	  }
  }
  else //// Indoor case
  {
    incar_loss = 0.;
  }


  ///////////////////////////// Building penetration loss /////////////////////////////////

  double glass_L    = 2 + 0.2 * Frequency / (1000000000);      ///// Standard multi-pane glass , Frequency = GHz
  double IRRglass_L = 23 + 0.3 * Frequency / (1000000000);     ///// Infrared Reflective(IRR) glass , Frequency = GHz
  double concrete_L = 5 + 4 * Frequency / (1000000000);        ///// Concrete glass , Frequency = GHz
  double wood_L     = 4.85 + 0.12 * Frequency / (1000000000);  ///// Wood glass , Frequency = GHz

  double indoor_L;

  if (TYPE == 12) // Dense Urban UMa
  {
	  if (ms[self_ms_idx].Indoor == 0)   ///// Outdoor
	  {
		  Otoi_loss = 0;
	  }
	  else                               /////// Indoor
	  {
		  if (Channel_Model_Type == 1)   ///// Model B
		  {
			  if (randnum.u() > 0.2) //// 80% Low Loss  O2I
			  {
				  /////// Low-loss model

				  Otoi_loss = 5 - 10 * log10(0.3 * pow(10, -1 * (glass_L / 10)) + 0.7 * pow(10, -1 * (concrete_L / 10)));
				  indoor_L = 0.5 * MIN((25 * randnum.u()), (25 * randnum.u()));

				  Otoi_loss = Otoi_loss + indoor_L + 4.4 * randnum.n();

			  }
			  else           ////// 20& High Loss O2I
			  {
				  /////// High-loss model

				  Otoi_loss = 5 - 10 * log10(0.7 * pow(10, -1 * (IRRglass_L / 10)) + 0.3 * pow(10, -1 * (concrete_L / 10)));
				  indoor_L = 0.5 * MIN((25 * randnum.u()), (25 * randnum.u()));

				  Otoi_loss = Otoi_loss + indoor_L + 6.5 * randnum.n();
			  }
		  }
		  else if (Channel_Model_Type == 0)  //// Model A 
		  {
			  if (Frequency > 6000000000)  //// above 6GHz
			  {
				  if (randnum.u() > 0.2) //// 80% Low Loss  O2I
				  {
					  /////// Low-loss model

					  Otoi_loss = 5 - 10 * log10(0.3 * pow(10, -1 * (glass_L / 10)) + 0.7 * pow(10, -1 * (concrete_L / 10)));
					  indoor_L = 0.5 * MIN((25 * randnum.u()), (25 * randnum.u()));

					  Otoi_loss = Otoi_loss + indoor_L + 4.4 * randnum.n();
				  }
				  else
				  {
					  /////// High-loss model

					  Otoi_loss = 5 - 10 * log10(0.7 * pow(10, -1 * (IRRglass_L / 10)) + 0.3 * pow(10, -1 * (concrete_L / 10)));
					  indoor_L = 0.5 * MIN((25 * randnum.u()), (25 * randnum.u()));

					  Otoi_loss = Otoi_loss + indoor_L + 6.5 * randnum.n();
				  }
			  }
			  else                         //// below 6GHz
			  {
				  Otoi_loss = 20;
				  indoor_L = 0.5 * (25 * randnum.u());

				  Otoi_loss = Otoi_loss + indoor_L;
			  }
		  }
		  else
		  {
			  cout << "Channel Model Type ERROR!! - Otoi Loss" << endl;
		  }
	  }
  }
  else if (TYPE == 13) // Rural
  {
	  if (ms[self_ms_idx].Indoor == 0)   ///// Outdoor
	  {
		  Otoi_loss = 0;
	  }
	  else                               /////// Indoor
	  {
		  if (Channel_Model_Type == 1)   ///// Model B
		  {
			  /////// 100% Low-loss model
			  Otoi_loss = 5 - 10 * log10(0.3 * pow(10, -1 * (glass_L / 10)) + 0.7 * pow(10, -1 * (concrete_L / 10)));
			  indoor_L = 0.5 * MIN((10 * randnum.u()), (10 * randnum.u()));

			  Otoi_loss = Otoi_loss + indoor_L + 4.4 * randnum.n();

		  }
		  else if (Channel_Model_Type == 0)  //// Model A 
		  {
			  if (Frequency > 6000000000)  //// above 6GHz
			  {
				  /////// 100% Low-loss model
				  Otoi_loss = 5 - 10 * log10(0.3 * pow(10, -1 * (glass_L / 10)) + 0.7 * pow(10, -1 * (concrete_L / 10)));
				  indoor_L = 0.5 * MIN((10 * randnum.u()), (10 * randnum.u()));

				  Otoi_loss = Otoi_loss + indoor_L + 4.4 * randnum.n();
			  }
			  else                         //// below 6GHz
			  {
				  Otoi_loss = 10;
				  indoor_L = 0.5 * (10 * randnum.u());

				  Otoi_loss = Otoi_loss + indoor_L;
			  }
		  }
		  else
		  {
			  cout << "Channel Model Type ERROR!! - Otoi Loss" << endl;
		  }
	  }
  }
  else   ///// Indoor case
  {
	  Otoi_loss = 0.;
  }


  // initial setting
  _throughput = 0 ;
  CQI_offset = 0;
  num_added = 0;
  num_re_tx =    0;
  ACK = true;

  //H_m = NULL;
  //HI_m = NULL;
}


double LINK::Get_angle(LOCATION interferer, LOCATION victim, LOCATION interfere_rx  )
{
	double angle = 0;
	double vector1_x, vector1_y, vector2_x, vector2_y;
	double acos_input = 0;
	

	vector1_x = victim.x - interferer.x;
	vector1_y = victim.y - interferer.y;
	vector2_x = interfere_rx.x - interferer.x;
	vector2_y = interfere_rx.y - interferer.y;
	acos_input = (vector1_x*vector2_x + vector1_y*vector2_y)
		/ (sqrt(vector1_x*vector1_x + vector1_y*vector1_y)*sqrt(vector2_x*vector2_x + vector2_y*vector2_y));

	if (acos_input >= 1)
	{
		angle = 0;
	}
	else
	{
		angle = acos(acos_input);
		if ((victim.y-interfere_rx.y)<0.)
		{
			angle = -angle;
		}

	}

	return angle * 180. / pi;

}


double LINK::Get_antgain(LOCATION interferer,LOCATION victim,int sector_idx)
{
	LOCATION vector;

	vector.x = interferer.x + cos((30 + 120 * (sector_idx - 1))*pi / 180.);
	vector.y = interferer.y + sin((30 + 120 * (sector_idx - 1))*pi / 180.);
		
	if (TYPE == 11)   /// Indoor
	{

		double distance = sqrt((interferer.x - victim.x)*(interferer.x - victim.x) + (interferer.y - victim.y)*(interferer.y - victim.y));
		double h_angle = 0.;    ///// ϴ, 3-Sector  Ƿ, horizontal gain 0 
		//double v_angle = atan2((bs_height - ms_height), distance)*180. / pi + 90.;
		double v_angle = atan2((bs_height - ms_height), distance)*180. / pi + 90.;

		double h_antgain = 0.;   ///// ϴ, 3-Sector  Ƿ, horizontal gain 0 
		double v_antgain = -MIN(12. * ((v_angle - 90) / 90.) * ((v_angle - 90) / 90.), 25.);

		return max_antgain - MIN(-(h_antgain + v_antgain), 25.);
		//return -MIN(-(h_antgain + v_antgain), 25.);
	}
	else if (TYPE == 12 || TYPE == 13)   ////// Dense Urban & Rural
	{

		double ms_height_in_channel = 0;

		ms_height_in_channel = ms[self_ms_idx].MS_HEIGHT_FINAL;

		double distance = sqrt((interferer.x - victim.x)*(interferer.x - victim.x) + (interferer.y - victim.y)*(interferer.y - victim.y));
		double h_angle = Get_angle(interferer, victim, vector);
		double v_angle = atan2((bs_height - ms_height_in_channel), distance)*180. / pi + 90;


		double theta_prime = Get_LCS_theta(30, 90, 0, v_angle, h_angle);
		double pi_prime = Get_LCS_pi(30, 90, 0, v_angle, h_angle);

		/*
		cout << "h_angle = " << h_angle << endl;
		cout << "v_angle = " << v_angle << endl;
		cout << "theta prime = " << theta_prime << endl;
		cout << "pi prime = " << pi_prime << endl;
		getchar();
		*/
		/*
		double h_antgain = -MIN(12. * (h_angle / 65.)*(h_angle / 65.), 30.);
		double v_antgain = -MIN(12. * ((v_angle - 90) / 65.) * ((v_angle - 90) / 65.), 30.);
		*/

		double h_antgain = -MIN(12. * (pi_prime / 65.)*(pi_prime / 65.), 30.);
		double v_antgain = -MIN(12. * ((theta_prime - 90) / 65.) * ((theta_prime - 90) / 65.), 30.);

		return max_antgain - MIN(-(h_antgain + v_antgain), 30.);

	}

	else
	{
		return 0;
	}

}


double Get_LCS_theta(double alpha, double beta, double gamma, double GCS_theta, double GCS_pi)
{
	return acos( cos(beta)*cos(gamma)*cos(GCS_theta) + (sin(beta)*cos(gamma)*cos(GCS_pi-alpha) - sin(gamma)*sin(GCS_pi-alpha))*sin(GCS_theta)) * 180./pi;
}

double Get_LCS_pi(double alpha, double beta, double gamma, double GCS_theta, double GCS_pi)
{

	double real;
	double imag;

	real = (cos(beta)*sin(GCS_theta)*cos(GCS_pi - alpha) - sin(beta)*cos(GCS_theta));
	imag = (cos(beta)*sin(gamma)*cos(GCS_theta) + (sin(beta)*sin(gamma)*cos(GCS_pi - alpha) + cos(gamma)*sin(GCS_pi - alpha))*sin(GCS_theta));

	complex <double> result = (real, imag);

	return arg(result) * 180./pi;

}