#include "common.h"

#include <Eigen/LU>

/*===================================================================
FUNCTION: Sector::Scheduling( int )
===================================================================*/
void Sector::Scheduling(int _Scheduling_Type)
{

	int Scheduling_Type_S = _Scheduling_Type;
	
	if (Scheduling_Type_S == 0)

	{
		Scheduling_RoundRobin();
	}

	else if (Scheduling_Type_S == 1)
	{
		//Scheduling_Proportional_Fair();
		cout << "Scheduling_Type input error, Use RoundRobin scheduling......" << endl;
		Scheduling_RoundRobin();

	}

	else
	{
		cout << "Scheduling_Type input error, Use RoundRobin scheduling......" << endl;
		Scheduling_RoundRobin();
	}

}


/*===================================================================
FUNCTION: Sector::Scheduling_RoundRobin( int )
===================================================================*/
void Sector::Scheduling_RoundRobin(void)
{
	/*===================================================================
	FUNCTION : Read_Ch_Feedback()
	INPUT    : pppPMI_map,pppCQI_map
	OUTPUT   : PMI_read, CQI_read, MCS_decision,CQI_AVR,METRIC
	===================================================================*/
	Read_Ch_Feedback();

	/*===================================================================
	FUNCTION : Initialize_schedule_map()
	ppschedulewrite[rb_idx][stream].ue_selected
	ppschedulewrite[rb_idx][stream].mcs_selected
	===================================================================*/
	Initialize_schedule_map();

	////////////////////////////////////////////////////////////////////
	if (W == NULL)
	{
		W = new MatrixXcd[num_rbs_in_1ms];

		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			MatrixXcd temp = MatrixXcd::Zero(4, 2);
			W[freq_idx] = temp;
		}
	}
	////////////////////////////////////////////////////////////////////

	int selected_ue = (int)(ue_in_control.size() * randnum.u()); //// one user has 50 RB (when num_TX == 1) 

	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		for (int stream = 0; stream < NUM_TX; stream++)
		{
			ppschedulewrite[rb_idx][stream].ue_selected = ue_in_control[selected_ue];
		}

	}

	/*===================================================================
	FUNCTION : Set_AVR_Cqi()
	INPUT    : CQI_read
	OUTPUT   : ppschedulewrite[freq_idx][stream].mcs_selected
	===================================================================*/
	Set_AVR_Cqi();
}


void Sector::Read_Ch_Feedback(void)
{

	if (t < N_pf + SCHEDULE_DELAY)
	{
		_N_pf = (t + 1) - SCHEDULE_DELAY;
	}
	else
	{
		_N_pf = N_pf;
	}

	double alpha = 1;


	if (PMI_read == NULL)
	{
		PMI_read = new int    *[ue_in_control.size()];
		CQI_read = new double *[ue_in_control.size()];
		MCS_decision = new int    *[ue_in_control.size()];
		CQIndex_decision = new int    *[ue_in_control.size()]; // 
		CQI_AVR = new double *[ue_in_control.size()];
		METRIC = new double *[ue_in_control.size()];

		for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
		{
			PMI_read[ue_idx] = new int[num_rbs_in_1ms];
			CQI_read[ue_idx] = new double[num_rbs_in_1ms];
			MCS_decision[ue_idx] = new int[num_rbs_in_1ms];
			CQIndex_decision[ue_idx] = new int[num_rbs_in_1ms];
			CQI_AVR[ue_idx] = new double[num_rbs_in_1ms];
			METRIC[ue_idx] = new double[num_rbs_in_1ms];
		}

		for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				PMI_read[ue_idx][rb_idx] = 0;
				CQI_read[ue_idx][rb_idx] = 0;
				MCS_decision[ue_idx][rb_idx] = 0;
				CQIndex_decision[ue_idx][rb_idx] = 0;
				CQI_AVR[ue_idx][rb_idx] = 0;
				METRIC[ue_idx][rb_idx] = 0;
			}
		}
	}

	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
	{
		for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
		{
			PMI_read[ue_idx][rb_idx] = pppPMI_map[ue_in_control[ue_idx]][rb_idx][(cqi_history_length + t - SCHEDULE_DELAY) % cqi_history_length];///Ͱ PMIʿ PMI о´.
			CQI_read[ue_idx][rb_idx] = pppCQI_Map[ue_in_control[ue_idx]][rb_idx][(cqi_history_length + t - SCHEDULE_DELAY) % cqi_history_length]; ///Ͱ CQIʿ CQI о´.

			double cqi_offset_added = CQI_read[ue_idx][rb_idx];

			int mcs_up = determine_MCS(linear2dB(cqi_offset_added));//+ mcs_offset; /// 
			int cqi_up = determine_CQI(linear2dB(cqi_offset_added));//+ mcs_offset; /// 

			if (true)  // Use MCS index
			{
				if (mcs_up > 27)
				{
					MCS_decision[ue_idx][rb_idx] = 27;
				}
				else if (mcs_up <0)
				{
					MCS_decision[ue_idx][rb_idx] = 0;
				}
				else
				{
					MCS_decision[ue_idx][rb_idx] = mcs_up;
				}

			}
			else  ///// Use CQI index 
			{
				// CQI index ׳ 
				MCS_decision[ue_idx][rb_idx] = mcs_up;

			}




			if (cqi_up > 15)
			{
				CQIndex_decision[ue_idx][rb_idx] = 15;
			}
			else if (cqi_up < 1)
			{
				CQIndex_decision[ue_idx][rb_idx] = 1;
			}
			else
			{
				CQIndex_decision[ue_idx][rb_idx] = cqi_up;
			}


		}

	}


	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++) 
	{
		for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
		{


			if (CQI_AVR[ue_idx][rb_idx])
			{
				CQI_AVR[ue_idx][rb_idx] = (1. / _N_pf) * log2(1 + CQI_read[ue_idx][rb_idx]) + (1 - 1. / _N_pf)* CQI_AVR[ue_idx][rb_idx];
			}
			else  ///// first loop 
			{
				CQI_AVR[ue_idx][rb_idx] = log2(1 + CQI_read[ue_idx][rb_idx]);

			}
		}
	}

	if (Scheduling_Type == 0)
	{
		for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				METRIC[ue_idx][rb_idx] = log2(1 + CQI_read[ue_idx][rb_idx]);//
			}
		}
	}
	else if (Scheduling_Type == 1)  // 
	{
		for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
			{
				METRIC[ue_idx][rb_idx] = log2(1 + CQI_read[ue_idx][rb_idx]) / pow(CQI_AVR[ue_idx][rb_idx], alpha);
			}
		}
	}
}


int Sector::determine_MCS(double sinr_estimate)
{


	if (false)  /// CQI
	{
		int _CQI_decision = 0;

		/////  'MCS Selection for Throughput Improvement in Downlink LTE Systems - Jiancun Fan, Qinye Yin, Geoffrey Ye Li, Bingguang Peng, and Xiaolong Zhu'  Table 1 SINR Threshold .... vienna  Thresholdΰ 

		if (sinr_estimate < -9.478)  { _CQI_decision = 1; }
		else if (sinr_estimate < -6.658)  { _CQI_decision = 2; }
		else if (sinr_estimate < -4.098)  { _CQI_decision = 3; }
		else if (sinr_estimate < -1.798)  { _CQI_decision = 4; }
		else if (sinr_estimate < 0.399)   { _CQI_decision = 5; }
		else if (sinr_estimate < 2.424)   { _CQI_decision = 6; }
		else if (sinr_estimate < 4.489)   { _CQI_decision = 7; }
		else if (sinr_estimate < 6.367)   { _CQI_decision = 8; }
		else if (sinr_estimate < 8.456)   { _CQI_decision = 9; }
		else if (sinr_estimate < 10.266)  { _CQI_decision = 10; }
		else if (sinr_estimate < 12.218)  { _CQI_decision = 11; }
		else if (sinr_estimate < 14.122)  { _CQI_decision = 12; }
		else if (sinr_estimate < 15.849)  { _CQI_decision = 13; }
		else if (sinr_estimate < 17.786)  { _CQI_decision = 14; }
		else                              { _CQI_decision = 15; }


		return _CQI_decision;

	}

	else if (true)  //// CQI index   /////// ϰ MCS index Ѵ
	{
		int _CQI_decision = 0;

		int _MCS_decision = 0;


		/////  'MCS Selection for Throughput Improvement in Downlink LTE Systems - Jiancun Fan, Qinye Yin, Geoffrey Ye Li, Bingguang Peng, and Xiaolong Zhu'  Table 1 SINR Threshold .... vienna  Thresholdΰ 

		/*
			if      (sinr_estimate < -9.478)  {_CQI_decision = 1;}
			else if (sinr_estimate < -6.658)  {_CQI_decision = 2;}
			else if (sinr_estimate < -4.098)  {_CQI_decision = 3;}
			else if (sinr_estimate < -1.798)  {_CQI_decision = 4;}
			else if (sinr_estimate < 0.399)   {_CQI_decision = 5;}
			else if (sinr_estimate < 2.424)   {_CQI_decision = 6;}
			else if (sinr_estimate < 4.489)   {_CQI_decision = 7;}
			else if (sinr_estimate < 6.367)   {_CQI_decision = 8;}
			else if (sinr_estimate < 8.456)   {_CQI_decision = 9;}
			else if (sinr_estimate < 10.266)  {_CQI_decision = 10;}
			else if (sinr_estimate < 12.218)  {_CQI_decision = 11;}
			else if (sinr_estimate < 14.122)  {_CQI_decision = 12;}
			else if (sinr_estimate < 15.849)  {_CQI_decision = 13;}
			else if (sinr_estimate < 17.786)  {_CQI_decision = 14;}
			else                              {_CQI_decision = 15;}
		*/


			/// 񿣳 BLER Ŀ  ϴ°
		if (sinr_estimate > SINR_threshold_dB[14])       { _CQI_decision = 15; }
		else if (sinr_estimate > SINR_threshold_dB[13])  { _CQI_decision = 14; }
		else if (sinr_estimate > SINR_threshold_dB[12])  { _CQI_decision = 13; }
		else if (sinr_estimate > SINR_threshold_dB[11])  { _CQI_decision = 12; }
		else if (sinr_estimate > SINR_threshold_dB[10])  { _CQI_decision = 11; }
		else if (sinr_estimate > SINR_threshold_dB[9])  { _CQI_decision = 10; }
		else if (sinr_estimate > SINR_threshold_dB[8])  { _CQI_decision = 9; }
		else if (sinr_estimate > SINR_threshold_dB[7])  { _CQI_decision = 8; }
		else if (sinr_estimate > SINR_threshold_dB[6])  { _CQI_decision = 7; }
		else if (sinr_estimate > SINR_threshold_dB[5])  { _CQI_decision = 6; }
		else if (sinr_estimate > SINR_threshold_dB[4])  { _CQI_decision = 5; }
		else if (sinr_estimate > SINR_threshold_dB[3])  { _CQI_decision = 4; }
		else if (sinr_estimate > SINR_threshold_dB[2])  { _CQI_decision = 3; }
		else if (sinr_estimate > SINR_threshold_dB[1])  { _CQI_decision = 2; }
		else if (sinr_estimate > SINR_threshold_dB[0])  { _CQI_decision = 1; }
		else                                            { _CQI_decision = 1; }



			//_CQI_decision = 8;



			// TS 36.101  TABLE A.4-7  MCS Scheme 1 
			if      (_CQI_decision == 1)   { _MCS_decision = 0;  }
			else if (_CQI_decision == 2)   { _MCS_decision = 0;  }
			else if (_CQI_decision == 3)   { _MCS_decision = 2;  }
			else if (_CQI_decision == 4)   { _MCS_decision = 4;  }
			else if (_CQI_decision == 5)   { _MCS_decision = 6;  }
			else if (_CQI_decision == 6)   { _MCS_decision = 8;  }
			else if (_CQI_decision == 7)   { _MCS_decision = 11; }
			else if (_CQI_decision == 8)   { _MCS_decision = 13; }
			else if (_CQI_decision == 9)   { _MCS_decision = 16; }
			else if (_CQI_decision == 10)  { _MCS_decision = 18; }
			else if (_CQI_decision == 11)  { _MCS_decision = 21; }
			else if (_CQI_decision == 12)  { _MCS_decision = 23; }
			else if (_CQI_decision == 13)  { _MCS_decision = 25; }
			else if (_CQI_decision == 14)  { _MCS_decision = 27; }
			else if (_CQI_decision == 15)  { _MCS_decision = 27; }

			/*
			cout << "SINR = " << sinr_estimate << endl;
			cout << "CQI = " << _CQI_decision << endl;
			cout << "MCS = " << _MCS_decision << endl;
			getchar();
			*/

			return _MCS_decision;

	
	}


	else  //   ڵ 
	{
		int _MCS_decision = 0;



	//	cout << "sinr_estimate dB = " << sinr_estimate << endl;


		for (int threshold_idx = 0; threshold_idx < NUM_CQI_MCS_TYPES - 1; threshold_idx++)
		{

			if (sinr_estimate < Threshold[0])
			{
				_MCS_decision = 0;
			}
			else if (sinr_estimate >= Threshold[threshold_idx] && sinr_estimate  < Threshold[threshold_idx + 1])
			{
				_MCS_decision = threshold_idx;
			}
			else if (sinr_estimate > Threshold[NUM_CQI_MCS_TYPES - 1])
			{
				_MCS_decision = NUM_CQI_MCS_TYPES - 1;
			}



		}


		//cout << "_MCS_decision idx = " << _MCS_decision << endl;

		//getchar();



		return _MCS_decision;

	}


	
}





int Sector::determine_CQI(double sinr_estimate)
{

	if (true)  /// CQI
	{
		int _CQI_decision = 0;

		/////  'MCS Selection for Throughput Improvement in Downlink LTE Systems - Jiancun Fan, Qinye Yin, Geoffrey Ye Li, Bingguang Peng, and Xiaolong Zhu'  Table 1 SINR Threshold .... vienna  Thresholdΰ 

		/*
		if (sinr_estimate < -9.478)  { _CQI_decision = 1; }
		else if (sinr_estimate < -6.658)  { _CQI_decision = 2; }
		else if (sinr_estimate < -4.098)  { _CQI_decision = 3; }
		else if (sinr_estimate < -1.798)  { _CQI_decision = 4; }
		else if (sinr_estimate < 0.399)   { _CQI_decision = 5; }
		else if (sinr_estimate < 2.424)   { _CQI_decision = 6; }
		else if (sinr_estimate < 4.489)   { _CQI_decision = 7; }
		else if (sinr_estimate < 6.367)   { _CQI_decision = 8; }
		else if (sinr_estimate < 8.456)   { _CQI_decision = 9; }
		else if (sinr_estimate < 10.266)  { _CQI_decision = 10; }
		else if (sinr_estimate < 12.218)  { _CQI_decision = 11; }
		else if (sinr_estimate < 14.122)  { _CQI_decision = 12; }
		else if (sinr_estimate < 15.849)  { _CQI_decision = 13; }
		else if (sinr_estimate < 17.786)  { _CQI_decision = 14; }
		else                              { _CQI_decision = 15; }


		return _CQI_decision;
		*/


			/*  񿣳 BLER Ŀ꿡  CQI bler 0.1   SNR dB
			CQI 1 = -6.3
			-5.725
			CQI 2 = -5.15
			-3.975
			CQI 3 = -2.8
			-2.025
			CQI 4 = -1.25
			-0.25
			CQI 5 = 0.75
			1.875
			CQI 6 = 3
			3.85
			CQI 7 = 4.7
			5.625
			CQI 8 = 6.55
			7.55
			CQI 9 = 8.55
			9.45
			CQI 10 = 10.35
			11.325
			CQI 11 = 12.3
			13.225
			CQI 12 = 14.15
			15.025
			CQI 13 = 15.9
			16.85
			CQI 14 = 17.8
			18.825
			CQI 15 = 19.85
			*/

		if (sinr_estimate > SINR_threshold_dB[14])       { _CQI_decision = 15; }
		else if (sinr_estimate > SINR_threshold_dB[13])  { _CQI_decision = 14; }
		else if (sinr_estimate > SINR_threshold_dB[12])  { _CQI_decision = 13; }
		else if (sinr_estimate > SINR_threshold_dB[11])  { _CQI_decision = 12; }
		else if (sinr_estimate > SINR_threshold_dB[10])  { _CQI_decision = 11; }
		else if (sinr_estimate > SINR_threshold_dB[9])  { _CQI_decision = 10; }
		else if (sinr_estimate > SINR_threshold_dB[8])  { _CQI_decision = 9; }
		else if (sinr_estimate > SINR_threshold_dB[7])  { _CQI_decision = 8; }
		else if (sinr_estimate > SINR_threshold_dB[6])  { _CQI_decision = 7; }
		else if (sinr_estimate > SINR_threshold_dB[5])  { _CQI_decision = 6; }
		else if (sinr_estimate > SINR_threshold_dB[4])  { _CQI_decision = 5; }
		else if (sinr_estimate > SINR_threshold_dB[3])  { _CQI_decision = 4; }
		else if (sinr_estimate > SINR_threshold_dB[2])  { _CQI_decision = 3; }
		else if (sinr_estimate > SINR_threshold_dB[1])  { _CQI_decision = 2; }
		else if (sinr_estimate > SINR_threshold_dB[0])  { _CQI_decision = 1; }
		else                                            { _CQI_decision = 1; }

		return _CQI_decision;




	}
}


/*===================================================================
FUNCTION: Sector::Initialize_schedule_map( int )
===================================================================*/
void Sector::Initialize_schedule_map(void)
{
	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		for (int stream = 0; stream < NUM_TX; stream++)
		{
			ppschedulewrite[rb_idx][stream].ue_selected = -1;
			ppschedulewrite[rb_idx][stream].mcs_selected = -1; // intitialize
			ppschedulewrite[rb_idx][stream].cqi_selected = -1; // intitialize
		}
	}

}



/*===================================================================
FUNCTION: Sector::Set_AVR_Cqi( void )
DESCRIPTION: Check the cqi for every allocated resource block, and Set it with minimum Cqi.
===================================================================*/

void Sector::Set_AVR_Cqi(void)   ////
{
	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
	{
		int count = 0;
		double sum = 0;
		double e_sinr[15] = { 0. };
		int decision_CQI_index = 0;
		
		int determined_cqi;
		int determined_mcs;
		double avr_cqi;

		///////////////////////////////////////
		//// Average SINR
		//////////////////////////////////////
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int stream = 0; stream < NUM_TX; stream++)
			{
				if (ppschedulewrite[freq_idx][stream].ue_selected == ue_in_control[ue_idx])
				{
					sum += CQI_read[ue_idx][freq_idx];
					count++;
				}
			}
		}


		avr_cqi = sum / (double)count;

		determined_mcs = determine_MCS(linear2dB(avr_cqi));// + 3;
		determined_cqi = determine_CQI(linear2dB(avr_cqi));// + 3;

		if (determined_mcs > 27)
		{
			determined_mcs = 27;
		}

		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int stream = 0; stream < NUM_TX; stream++)
			{
				if (ppschedulewrite[freq_idx][stream].ue_selected == ue_in_control[ue_idx])
				{
					ppschedulewrite[freq_idx][stream].mcs_selected = determined_mcs;
					ppschedulewrite[freq_idx][stream].cqi_selected = determined_cqi;
				}
			}
		}
	}
}




/*===================================================================
FUNCTION: Sector::Initial_Setting( int )

DESCRIPTION:
===================================================================*/
void Sector::Initial_Setting(int rb_idx)
{
	scheduled_ue.clear();
	selected_ue.clear();
	prev_sum_SINR = -pow(10., 10.);
	sum_of_SINR = 0.;

	for (int stream = 0; stream < NUM_TX; stream++)
	{
		ppschedulewrite[rb_idx][stream].ue_selected = -1;
	}
}


void Sector::first_max_ue(int rb_idx)
{
	selected_ue.clear();

	double MAX_SINR = -pow(10., 10.);
	int    MAX_UE = -1;

	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
	{
		if (METRIC[ue_idx][rb_idx] > MAX_SINR &&  METRIC[ue_idx][rb_idx] < prev_max_sinr)  //// տ prev_max_sinr = pow(10., 10.) 
		{
			MAX_SINR = METRIC[ue_idx][rb_idx];
			MAX_UE = ue_idx;
		}
	}
	selected_ue.push_back(MAX_UE);  //// SINR  ū ue ϳ  , selected_ue size 1ϰ
	prev_max_sinr = MAX_SINR;

	//cout << "selected_ue[0] = " << selected_ue[0] << endl;


}

/*===================================================================
FUNCTION: Sector::SUS_procedure( int )

DESCRIPTION: Sector object gets to know the index of itself
===================================================================*/
void Sector::SUS_procedure(int rb_idx)
{


	/////
	// SUS = semi orthogonal User selection
	////
	

	//// ó selected_ue.size() 1, տ max sinr  ue idx ϳ... 

	while (1) //selected_ue.size() != num_transmit_antenna
	{

		//cout << "SUS start" << endl;



		if (selected_ue[0] == -1)   // selected ue size 2϶  ͼ 
		{
			//cout << "BREAK!" << endl;

			//cout << "scheduled_ue size = " << scheduled_ue.size() << endl;
			//cout << "scheduled_ue[0] = " << scheduled_ue[0] << endl;
			//cout << "scheduled_ue[1] = " << scheduled_ue[1] << endl;
			break;
		}

		if (selected_ue[selected_ue.size() - 1] > -1)
		{



			Get_User_Set(rb_idx);   ///// ue_in_control ߿, տ õ ue semi orthogonal Ǵ user ãƼ user_set ִ´.

			//cout << "user set = " << user_set.size() << endl;
			//getchar();


			double MAX_SINR = -500;
			int    MAX_UE = -1;
			for (int j = 0; j < (int)user_set.size(); j++)
			{
				if (METRIC[user_set[j]][rb_idx] > MAX_SINR)
				{
					MAX_SINR = METRIC[user_set[j]][rb_idx];
					MAX_UE = user_set[j];
				}
			}
			selected_ue.push_back(MAX_UE);

			//// user set߿ sinr   ϳ ؼ selected_ue ִ´ -> ̷ ϸ selected ue 2 ε  
			// ι° selected ue -1   user_set size 0ϋ... semi orthogonal ε


		}
		else
		{
			cout << "push back -1!!" << endl; //  δ ȵ, ȵ° ´µ
			getchar();
			selected_ue.push_back(-1);
		}

		//cout << "selected ue size = " << selected_ue.size() << endl;
		//cout << "selected ue[1] = " << selected_ue[1] << endl;
		//getchar();

		if (selected_ue.size() == mx_ue_mumimo || selected_ue.size() == ue_in_control.size()) //num_transmit_antenna    /// mx_ue_mumimo = 2  configuration ϰ, 4x2 mu mimo̸ 2 Ǵ° ´µ
		{

			for (int i = 0; i < (int)selected_ue.size(); i++)
			{
				if (selected_ue[i] > -1)
				{
					sum_of_SINR += METRIC[selected_ue[i]][rb_idx];
				}
			}


			//cout << "sum_of_SINR = " << sum_of_SINR << endl;
			//cout << "prev_sum_SINR = " << prev_sum_SINR << endl;
			//getchar();

			if (sum_of_SINR > prev_sum_SINR + 0.000000009)
			{
				scheduled_ue = selected_ue;
				prev_sum_SINR = sum_of_SINR;
			}
			sum_of_SINR = 0;

			first_max_ue(rb_idx);
		}
	}
}


void Sector::Get_User_Set(int rb_idx)
{
	user_set.clear();
	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
	{
		bool already_selected = false;

		for (int _ue_idx = 0; _ue_idx < (int)selected_ue.size(); _ue_idx++)
		{
			if (ue_idx == selected_ue[_ue_idx])
			{
				already_selected = true;
			}
		} // ̹ õ ue Ǵ


		bool semi_orthogonal = false;


		// ̹ õ ue ƴ϶...
		if (!already_selected)
		{
			semi_orthogonal = true;


			for (int j = 0; j < (int)selected_ue.size(); j++)
			{
				double MAX_Corr = -500;


				///  ڵ codeword rowvector ,  column vector Ǿ ִ.
				// Ʒó rowvector ؼ ־   ˾Ƽ  row ٲµ

				RowVector4cd A = code_word[PMI_read[selected_ue[j]][rb_idx]];
				RowVector4cd B = code_word[PMI_read[ue_idx][rb_idx]];



				double inner_prod = (A * B.adjoint()).norm();

			
				/*
				cout << "code_word[PMI_read[selected_ue[j]][rb_idx]] = " << code_word[PMI_read[selected_ue[j]][rb_idx]] << endl;
				cout << "code_word[PMI_read[ue_idx][rb_idx]] = " << code_word[PMI_read[ue_idx][rb_idx]] << endl;
				cout << "A vector = " << A << endl;
				cout << "B vector = " << B.adjoint() << endl;
				cout << "inner_prod = " << inner_prod << endl;
				getchar();
				*/



				if (inner_prod > epsilon)         //// epsilon ڵ忡 0.2   //// default 0.4 Ǿִµ..
				{
					semi_orthogonal = false;
				}
			}


			//////// Test  true غҴ. user_set Ʈ      ֵ... ///////////////
			//////////////////////////////
			semi_orthogonal = true;



			/*

			MatrixXcd Lambda = MatrixXcd::Zero(4, 4);


			/// Lambda 2 x 4ΰ .
			/// selected ue row 1, ueincontrol ue 1 ؼ semi orthogonal ϴµ

			//Lambda.clear();
			for (int i = 0; i < (int)selected_ue.size(); i++)
			{
				if (selected_ue[i] > -1)
				{
					RowVector4cd v_hat = code_word[PMI_read[selected_ue[i]][rb_idx]];

					//cout << "v_hat = " << v_hat << endl;

					v_hat(0) = complex <double>((int)(v_hat(0).real() * 1000)*(1. / 1000.), (int)(v_hat(0).imag() * 1000)*(1. / 1000.) );
					//v_hat(0).imag = (int)(v_hat(0).imag() * 1000)*(1. / 1000.);

					v_hat(1) = complex <double>((int)(v_hat(1).real() * 1000)*(1. / 1000.), (int)(v_hat(1).imag() * 1000)*(1. / 1000.));
					//v_hat(1).real = (int)(v_hat(1).real() * 1000)*(1. / 1000.);
					//v_hat(1).imag = (int)(v_hat(1).imag() * 1000)*(1. / 1000.);

					v_hat(2) = complex <double>((int)(v_hat(2).real() * 1000)*(1. / 1000.), (int)(v_hat(2).imag() * 1000)*(1. / 1000.));
					//v_hat(2).real = (int)(v_hat(2).real() * 1000)*(1. / 1000.);
					//v_hat(2).imag = (int)(v_hat(2).imag() * 1000)*(1. / 1000.);

					v_hat(3) = complex <double>((int)(v_hat(3).real() * 1000)*(1. / 1000.), (int)(v_hat(3).imag() * 1000)*(1. / 1000.));
					//v_hat(3).real = (int)(v_hat(3).real() * 1000)*(1. / 1000.);
					//v_hat(3).imag = (int)(v_hat(3).imag() * 1000)*(1. / 1000.);

					//Lambda.row(Lambda.rows() - 1) = v_hat;
					//Lambda.conservativeResize(Lambda.rows(), Lambda.cols() + 1);

					//cout << "v_hat(2) =" << v_hat(2) << endl;
					//getchar();

					Lambda.row(i) = v_hat;

					//cout << "Lambda = " << Lambda << endl;
					


				}
			}

			RowVector4cd v_hat = code_word[PMI_read[ue_idx][rb_idx]];
			v_hat(0) = complex <double>((int)(v_hat(0).real() * 1000)*(1. / 1000.), (int)(v_hat(0).imag() * 1000)*(1. / 1000.));
			//v_hat(0).imag = (int)(v_hat(0).imag() * 1000)*(1. / 1000.);

			v_hat(1) = complex <double>((int)(v_hat(1).real() * 1000)*(1. / 1000.), (int)(v_hat(1).imag() * 1000)*(1. / 1000.));
			//v_hat(1).real = (int)(v_hat(1).real() * 1000)*(1. / 1000.);
			//v_hat(1).imag = (int)(v_hat(1).imag() * 1000)*(1. / 1000.);

			v_hat(2) = complex <double>((int)(v_hat(2).real() * 1000)*(1. / 1000.), (int)(v_hat(2).imag() * 1000)*(1. / 1000.));
			//v_hat(2).real = (int)(v_hat(2).real() * 1000)*(1. / 1000.);
			//v_hat(2).imag = (int)(v_hat(2).imag() * 1000)*(1. / 1000.);

			v_hat(3) = complex <double>((int)(v_hat(3).real() * 1000)*(1. / 1000.), (int)(v_hat(3).imag() * 1000)*(1. / 1000.));
			//v_hat(3).real = (int)(v_hat(3).real() * 1000)*(1. / 1000.);
			//v_hat(3).imag = (int)(v_hat(3).imag() * 1000)*(1. / 1000.);

			//Lambda.row(Lambda.rows() - 1) = v_hat;
			//Lambda.conservativeResize(Lambda.rows(), Lambda.cols() + 1);
			Lambda.row(1) = v_hat;

			// Check What is the optimal value for the regularized channel inversion ?
			complex <double> _det = ((Lambda * Lambda.adjoint()).inverse()).determinant();


			cout << "_det = " << _det << endl;
			getchar();


			if (abs(_det) < 0.0001)
			{
				//cout << "_self_idx = " << _self_idx << endl;
				//cout << " rb_idx = " << rb_idx << endl;
				//cout << " _det = " << _det << endl;  getchar();
				semi_orthogonal = false;
			}

			*/

		}

		if (semi_orthogonal) //// semi ortho ϶ user set ue ε 
		{
			user_set.push_back(ue_idx);
		}
	}
}



/*===================================================================
FUNCTION: Sector::Transmit_Precoding( int )

DESCRIPTION:
===================================================================*/
void Sector::Transmit_Precoding(int rb_idx)
{
	MatrixXcd Lambda = MatrixXcd:: Zero(2,4);

	//Vector4cd Lambda;

	//Lambda.clear();

	Matrix2cd Identity = Matrix2cd::Identity();



	for (int i = 0; i < (int)scheduled_ue.size(); i++) /// scheduled_ue.size()  2
	{
		if (scheduled_ue[i] > -1)
		{



			RowVector4cd v_hat = code_word[PMI_read[scheduled_ue[i]][rb_idx]];

			//cout << "v_hat = " << v_hat << endl;


			v_hat(0) = complex <double>((int)(v_hat(0).real() * 1000)*(1. / 1000.), (int)(v_hat(0).imag() * 1000)*(1. / 1000.));
			//v_hat(0).imag = (int)(v_hat(0).imag() * 1000)*(1. / 1000.);

			v_hat(1) = complex <double>((int)(v_hat(1).real() * 1000)*(1. / 1000.), (int)(v_hat(1).imag() * 1000)*(1. / 1000.));
			//v_hat(1).real = (int)(v_hat(1).real() * 1000)*(1. / 1000.);
			//v_hat(1).imag = (int)(v_hat(1).imag() * 1000)*(1. / 1000.);

			v_hat(2) = complex <double>((int)(v_hat(2).real() * 1000)*(1. / 1000.), (int)(v_hat(2).imag() * 1000)*(1. / 1000.));
			//v_hat(2).real = (int)(v_hat(2).real() * 1000)*(1. / 1000.);
			//v_hat(2).imag = (int)(v_hat(2).imag() * 1000)*(1. / 1000.);

			v_hat(3) = complex <double>((int)(v_hat(3).real() * 1000)*(1. / 1000.), (int)(v_hat(3).imag() * 1000)*(1. / 1000.));
			//v_hat(3).real = (int)(v_hat(3).real() * 1000)*(1. / 1000.);
			//v_hat(3).imag = (int)(v_hat(3).imag() * 1000)*(1. / 1000.);




			
			//Lambda.row(Lambda.rows() - 1) = v_hat;
			//Lambda.conservativeResize(Lambda.rows(), Lambda.cols() + 1);

			Lambda.row(i) = v_hat;



		}
	}

	//cout << "Lambda = " << Lambda << endl;



	// Check What is the optimal value for the regularized channel inversion ?
	complex <double> _det = ((Lambda * Lambda.adjoint()).inverse()).determinant();
	////  ε 1 õ 쿡 inverse  ʾƼ ̻  Ե....

	///   Ȱ PMI϶ inverse 

	/// 0  Ͱ , Ʈ  ؼ 0   غ...........................


	//cout << "_det = " << _det << endl;
	//getchar();


	if (!abs(_det) || abs(_det) > pow(10., 10))
	{

		W[rb_idx] = Lambda.adjoint() * (0.01 * Identity + Lambda * Lambda.adjoint()).inverse();   /// ڵ忡 identity 4 Ǿִ. ̰ 4 Ƿ scheduled_ue 4 Ǿߵȴٴ¾ε, 4x2 mumimo̸ 2 ´° ƴѰ?, mx_ue_mimo 2 Ǿִµ.... 
	////  4x2            4x2                      2x2      2x4          4x2
	}
	else
	{

		W[rb_idx] = Lambda.adjoint() * (Lambda * Lambda.adjoint()).inverse();
	////  4x2            4x2             2x4          4x2
	}

	//cout << "W[rb_idx] = " << W[rb_idx] << endl;

	
	Vector2d sqrt_p;

	for (int i = 0; i < 2; i++)  /// ̰͵ Į  4 Ǿִµ 2 ٲ㼭 غ
	{
		sqrt_p(i) = (W[rb_idx].col(i)).norm();
	}

	//cout << "sqrt_p = " << sqrt_p << endl;

	W[rb_idx] = W[rb_idx] * sqrt_p.asDiagonal();

	//cout << "After diagonal W[rb_idx] = " << W[rb_idx] << endl;
	//getchar();
	
	/*
	for (int row = 0; row < W[rb_idx].rows(); row++)
	{
		for (int col = 0; col < W[rb_idx].cols(); col++)
		{
			if (isnan(W[rb_idx](row, col).real()) || isnan(W[rb_idx](row, col).imag()))
			{
				cout << " scheduled_ue.size() = " << scheduled_ue.size() << endl;
				cout << " ue_in_control.size() = " << ue_in_control.size() << endl;
				cout << " W(" << row << "," << col << ") = " << W[rb_idx](row, col) << endl;
				cout << " Lambda = " << Lambda << endl;
				cout << " inv( 0.01 * eye_c(4) + Lambda * Lambda.H() ) = " << (0.01 * Identity + Lambda * Lambda.adjoint()).inverse() << endl;
				cout << (Lambda * Lambda.adjoint()).inverse() << endl;
				getchar();
			}
		}
	}
	*/
	
}


/*===================================================================
FUNCTION: Sector::Set_Min_Cqi( void )

AUTHOR: SJ, KR, DW

DESCRIPTION: Check the cqi for every allocated resource block, and Set it with minimum Cqi.

NOTES: Updated 6 OCT 08
===================================================================*/
void Sector::Set_Min_Cqi(void)
{
	for (int ue_idx = 0; ue_idx < (int)ue_in_control.size(); ue_idx++)
	{
		int min_cqi = NUM_CQI_MCS_TYPES;
		// MIN CQI within UE setting for each frequency
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int stream = 0; stream < NUM_TX; stream++)
			if (ppschedulewrite[freq_idx][stream].ue_selected == ue_in_control[ue_idx])
			{
				if (min_cqi > MCS_decision[ue_idx][freq_idx])
				{
					min_cqi = MCS_decision[ue_idx][freq_idx];
				}
			}
		}

		// Set MIN CQI Scheduling for each frequency
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int stream = 0; stream < NUM_TX; stream++)
			if (ppschedulewrite[freq_idx][stream].ue_selected == ue_in_control[ue_idx])
			{
				ppschedulewrite[freq_idx][stream].mcs_selected = min_cqi;
			}
		}
	}
}

